/*
** pork_buddy_list.c - Buddy list screen widget.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <pork.h>

static chtype blist_indicator(u_int32_t status) {
	chtype indicator = '*';

	switch (status) {
		case STATUS_AWAY:
			indicator |= BLIST_COLOR_STATUS_AWAY;
			break;

		case STATUS_ACTIVE:
			indicator |= BLIST_COLOR_STATUS_ACTIVE;
			break;

		case STATUS_IDLE:
			indicator |= BLIST_COLOR_STATUS_IDLE;
			break;
	}

	return (indicator);
}

static int blist_compare(void *l, void *r) {
	struct blist_line *line = r;

	return (l != line->data);
}

static void blist_draw_line(struct buddy_list *blist,
							struct blist_line *line,
							int status)
{
	chtype chstr[blist->cols];
	int len;
	int i;
	u_int32_t flags;
	char *p;

	if (line->line < 0)
		return;

	if (status == 0)
		flags = 0;
	else
		flags = BLIST_COLOR_CURSOR;

	len = blist->cols - 2;
	p = line->text;

	chstr[0] = ' ' | flags;

	for (i = 1 ; *p != '\0' && i < len ; i++)
		chstr[i] = *p++ | flags;

	if (line->type == TYPE_BUDDY) {
		struct buddy *buddy = line->data;

		chstr[2] = blist_indicator(buddy->status) | flags;
	}

	while (i < len)
		chstr[i++] = ' ' | flags;

	chstr[i] = 0;
	mvwputstr(blist->win, line->line, 1, chstr);
}

static void blist_draw_curs(struct buddy_list *blist, int status) {
	if (blist->cursor == NULL)
		return;

	blist_draw_line(blist, blist->cursor->data, status);
}

void blist_draw_buddy_label(struct buddy_list *blist, struct buddy *buddy) {
	char label[128];

	if (blist->show_idle) {
		int i = 0;
		char status_str[64];
		size_t len = sizeof(status_str);
		int ret;

		status_str[0] = '\0';
		if (buddy->idle_time > 0 || buddy->warn_level > 0) {
			ret = snprintf(status_str, len, "[");
			i += ret;
			len -= ret;

			if (buddy->idle_time > 0) {
				char idle_time[32];

				time_to_str(buddy->idle_time, idle_time, sizeof(idle_time));

				if (buddy->warn_level > 0)
					ret = snprintf(&status_str[i], len, "i: %s, ", idle_time);
				else
					ret = snprintf(&status_str[i], len, "i: %s", idle_time);

				len -= ret;
				i += ret;
			}

			if (buddy->warn_level > 0) {
				ret = snprintf(&status_str[i], len, "w: %u%%",
						buddy->warn_level);
				i += ret;
				len -= ret;
			}

			ret = snprintf(&status_str[i], len, "]");
			i += ret;
			len -= ret;
		}
		
		snprintf(label, sizeof(label), "   %s %*s",
			buddy->name, blist->cols - strlen(buddy->name) - 7, status_str);
	} else
		snprintf(label, sizeof(label), "   %s", buddy->name);

	free(buddy->blist_line->text);
	buddy->blist_line->text = xstrdup(label);

	if (buddy->blist_line->line > 0) {
		if (blist->cursor->data == buddy->blist_line)
			blist_draw_line(blist, buddy->blist_line, 1);
		else
			blist_draw_line(blist, buddy->blist_line, 0);

		blist->dirty = 1;
	}
}

void blist_draw_group_label(struct buddy_list *blist, struct group *group) {
	char expand_char = '-';
	char label[128];

	if (group->blist_line == NULL)
		return;

	if (group->blist_line->collapsed)
		expand_char = '+';

	snprintf(label, sizeof(label), "%c %s (%u/%u)",
		expand_char, group->name, group->num_online,
		group->num_members);

	free(group->blist_line->text);
	group->blist_line->text = xstrdup(label);

	if (group->blist_line->line > 0) {
		if (blist->cursor->data == group->blist_line)
			blist_draw_line(blist, group->blist_line, 1);
		else
			blist_draw_line(blist, group->blist_line, 0);
		blist->dirty = 1;
	}
}

int blist_init(struct aim_priv *priv) {
	int cols = BUDDY_DEFAULT_WIDTH;
	struct buddy_list *blist;
	WINDOW *win;
	dlist_t *cur;
	int line = 1;

	/*
	** It's already been initialized.
	*/

	if (priv->blist != NULL)
		return (-1);

	win = newwin(screen.rows - STATUS_ROWS, cols, 0, screen.cols - cols);
	if (win == NULL)
		return (-1);

	set_default_win_opts(win);
	
	blist = xcalloc(1, sizeof(*blist));
	priv->blist = blist;

	blist->rows = screen.rows - STATUS_ROWS;
	blist->cols = cols;
	blist->win = win;
	blist->dirty = 1;

	/* Add the buddy groups to the buddy list */
	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct group *gr = cur->data;
		struct blist_line *new_line;
		char label[128];

		new_line = xmalloc(sizeof(*new_line));
		new_line->type = TYPE_GROUP;
		new_line->data = gr;
		new_line->collapsed = 0;

		snprintf(label, sizeof(label), "- %s (0/%u)", gr->name,
			gr->num_members);

		new_line->text = xstrdup(label);

		if (line <= BLIST_LAST_ROW(blist)) {
			new_line->line = line++;
		} else
			new_line->line = -1;

		gr->blist_line = new_line;
		blist->lines = dlist_add_tail(blist->lines, new_line);

		/*
		** If either the top or the bottom of the list is NULL,
		** the list was empty.
		*/

		if (blist->top == NULL)
			blist->top = blist->lines;

		if (blist->bot == NULL)
			blist->bot = blist->lines;
		else
			blist->bot = blist->bot->next;

		cur = cur->next;
	}

	blist->cursor = blist->lines;
	blist_redraw(blist);
	blist_draw_curs(blist, 1);

	return (0);
}

void blist_add(struct buddy_list *blist, struct buddy *buddy) {
	dlist_t *cur;
	struct blist_line *new_line;
	int renumber_list = 1;

	cur = dlist_find(blist->lines, buddy->group, blist_compare);
	if (cur == NULL)
		return;

	new_line = xcalloc(1, sizeof(*new_line));
	new_line->type = TYPE_BUDDY;
	new_line->data = buddy;
	buddy->blist_line = new_line;

	blist_draw_buddy_label(blist, buddy);

	while (1) {
		struct blist_line *cur_line = cur->data;
		struct blist_line *line;
		struct buddy *cur_buddy;

		if (cur->next == NULL)
			goto add_line;

		line = cur->next->data;
		if (line->type == TYPE_GROUP)
			goto add_line;

		cur_buddy = line->data;
		if (buddy->refnum < cur_buddy->refnum)
			goto add_line;

		cur = cur->next;
		continue;

add_line:
		blist->lines = dlist_add_after(blist->lines, cur, new_line);

		if (cur_line->line < 0) {
			new_line->line = cur_line->line;

			if (new_line->line == -2) {
				struct blist_line *cursor_line;

				blist_draw_curs(blist, 0);
				blist_scroll_down(blist);

				cursor_line = blist->cursor->data;
				if (cursor_line->line == -2)
					blist->cursor = blist->cursor->next;

				renumber_list = 0;
				blist_draw_curs(blist, 1);

				blist->dirty = 1;
			}
		} else if (cur == blist->bot) {
			if (cur_line->line < BLIST_LAST_ROW(blist)) {
				new_line->line = cur_line->line + 1;
				blist->bot = blist->bot->next;
			} else
				new_line->line = -1;
		} else {
			struct blist_line *bot_line = blist->bot->data;

			if (bot_line->line == BLIST_LAST_ROW(blist)) {
				bot_line->line = -1;

				if (blist->cursor == blist->bot)
					blist->cursor = blist->bot->prev;

				blist->bot = blist->bot->prev;
			}

			new_line->line = cur_line->line + 1;
		}

		break;
	}

	blist_draw_group_label(blist, buddy->group);

	/*
	** Increment the line number of all the lines that are on the screen.
	** If the line just added is off the screen, nothing needs to be done.
	*/

	if (new_line->line > 0 && renumber_list != 0) {
		cur = cur->next;

		while (cur->next != NULL) {
			struct blist_line *line = cur->next->data;

			if (line->line < 0)
				break;

			line->line++;
			cur = cur->next;
		}

		blist_redraw(blist);
		blist_draw_curs(blist, 1);

		blist->dirty = 1;
	}
}

void blist_del(struct buddy_list *blist, struct buddy *buddy) {
	dlist_t *cur;
	struct blist_line *line;
	dlist_t *next_node;
	int line_num;

	cur = dlist_find(blist->lines, buddy, blist_compare);
	if (cur == NULL)
		return;

	line = cur->data;
	if (line->line < 0)
		goto done;

	if (blist->bot->next != NULL)
		blist->bot = blist->bot->next;
	else if (cur == blist->bot) {
		blist->bot = blist->bot->prev;

		if (cur == blist->top)
			blist_scroll_up(blist);

		goto done;
	}

	if (cur == blist->top)
		blist->top = blist->top->next;

	next_node = cur->next;
	while (next_node != NULL) {
		struct blist_line *next_line = next_node->data;

		/*
		** The first line that's offscreen at the bottom
		** scrolls back onto the screen.
		*/

		if (next_line->line == -1) {
			next_line->line = BLIST_LAST_ROW(blist);
			break;
		}

		next_line->line--;
		next_node = next_node->next;
	}

done:
	if (line->type == TYPE_GROUP) {
		struct group *gr = line->data;

		gr->blist_line = NULL;
	} else {
		struct buddy *b = line->data;

		b->blist_line = NULL;
	}

	if (cur == blist->cursor) {
		blist->cursor = blist->cursor->next;

		if (blist->cursor == NULL)
			blist->cursor = blist->bot;
	}

	line_num = line->line;

	free(line->text);
	free(line);

	blist->lines = dlist_remove(blist->lines, cur);

	blist_draw_group_label(blist, buddy->group);

	if (line_num == -2) {
		struct blist_line *cursor_line;

		blist_draw_curs(blist, 0);
		blist_scroll_up(blist);

		/*
		** If scrolling the list caused the cursor to move off the screen
		** (i.e. it was on the bottom row before the window was scrolled)
		** adjust the cursor so it's visible.
		*/

		cursor_line = blist->cursor->data;
		if (cursor_line->line == -1)
			blist->cursor = blist->bot;

		blist_draw_curs(blist, 1);
		blist->dirty = 1;
	}

	/*
	** Just to make sure..
	*/

	if (blist->cursor == NULL)
		blist->cursor = blist->bot;
	else {
		line = blist->cursor->data;
		if (line->line == -1)
			blist->cursor = blist->bot;
		else if (line->line == -2)
			blist->cursor = blist->top;
	}

	if (line_num >= 0) {
		blist_redraw(blist);
		blist_draw_curs(blist, 1);
		blist->dirty = 1;
	}
}

/*
** XXX: do these.
*/

void blist_add_group(struct buddy_list *blist, struct group *group) {
	(void) blist;
	(void) group;
}

void blist_del_group(struct buddy_list *blist, struct group *group) {
	(void) blist;
	(void) group;
}

void blist_collapse_group(struct buddy_list *blist, struct blist_line *group) {
	struct group *gr = group->data;
	dlist_t *cur;
	void (*list_func)(struct buddy_list *, struct buddy *);

	if (group->collapsed)
		list_func = blist_add;
	else
		list_func = blist_del;

	for (cur = gr->member_list ; cur != NULL ; cur = cur->next) {
		struct buddy *buddy = cur->data;

		if (buddy->status != STATUS_OFFLINE)
			list_func(blist, buddy);
	}

	group->collapsed = !group->collapsed;

	if (group->line != -1) {
		blist->dirty = 1;
		blist_draw_group_label(blist, gr);
		blist_redraw(blist);
		blist_draw_curs(blist, 1);
	}
}

/*
** This function toggles the visibility of the buddy list.
*/

int blist_toggle(struct buddy_list *blist, struct imwindow *imwindow) {
	u_int32_t new_width;

	if (imwindow->blist_visible) {
		wclear(blist->win);
		imwindow->blist_visible = 0;
		imwindow->active_binds = &screen.binds.main;
		new_width = imwindow->swindow.cols + blist->cols;
	} else {
		blist_redraw(blist);
		blist_draw_curs(blist, 1);
		imwindow->blist_visible = 1;
		new_width = imwindow->swindow.cols - blist->cols;
	}

	blist->dirty = 1;
	wclear(imwindow->swindow.win);
	if (imwindow->swindow.visible)
		wnoutrefresh(imwindow->swindow.win);

	swindow_resize(&imwindow->swindow, imwindow->swindow.rows, new_width);
	imwindow_refresh(imwindow);

	return (imwindow->blist_visible);
}

/*
** This function toggles the reporting of idle time for
** buddies.
*/

int blist_show_idle(struct buddy_list *blist) {
	dlist_t *cur;
	struct imwindow *imw = screen.cur_window->data;
	u_int32_t new_width;
	u_int32_t new_swidth;

	if (blist->show_idle) {
		blist->show_idle = 0;
		new_swidth = imw->swindow.cols + BUDDY_IDLE_WIDTH;
		new_width = blist->cols - BUDDY_IDLE_WIDTH;
	} else {
		blist->show_idle = 1;
		new_swidth = imw->swindow.cols - BUDDY_IDLE_WIDTH;
		new_width = blist->cols + BUDDY_IDLE_WIDTH;
	}

	blist->cols = new_width;
	wresize(blist->win, imw->swindow.rows, new_width);
	mvwin(blist->win, 0, screen.cols - new_width);

	cur = blist->lines;
	while (cur != NULL) {
		struct blist_line *line = cur->data;

		if (line->type == TYPE_BUDDY)
			blist_draw_buddy_label(blist, line->data);

		cur = cur->next;
	}

	cur = screen.window_list;
	do {
		struct imwindow *imwindow = cur->data;

		if (imwindow->blist_visible) {
			wclear(imwindow->swindow.win);
			wnoutrefresh(imwindow->swindow.win);
			swindow_resize(&imwindow->swindow,
				imwindow->swindow.rows, new_swidth);
		}

		cur = cur->next;
	} while (cur != screen.window_list);

	blist->dirty = 1;
	blist_redraw(blist);
	blist_draw_curs(blist, 1);
	imwindow_refresh(imw);

	return (blist->show_idle);
}

void blist_cursor_up(struct buddy_list *blist) {
	if (blist->cursor == NULL || blist->cursor->prev == NULL)
		return;

	blist_draw_curs(blist, 0);

	if (blist->cursor == blist->top)
		blist_scroll_up(blist);

	blist->cursor = blist->cursor->prev;
	blist_draw_curs(blist, 1);

	blist->dirty = 1;
}

void blist_cursor_down(struct buddy_list *blist) {
	if (blist->cursor == NULL || blist->cursor->next == NULL)
		return;

	blist_draw_curs(blist, 0);

	if (blist->cursor == blist->bot)
		blist_scroll_down(blist);

	blist->cursor = blist->cursor->next;
	blist_draw_curs(blist, 1);

	blist->dirty = 1;
}

void blist_draw_border(struct buddy_list *blist) {
	struct imwindow *imwindow = screen.cur_window->data;
	attr_t box_attr;

	if (imwindow->active_binds == &screen.binds.main)
		box_attr = BLIST_COLOR_BORDER_INACTIVE;
	else
		box_attr = BLIST_COLOR_BORDER_ACTIVE;

	wattron(blist->win, box_attr);
	box(blist->win, ACS_VLINE, ACS_HLINE);
	wattroff(blist->win, box_attr);
}

void blist_scroll_down(struct buddy_list *blist) {
	struct blist_line *line = blist->top->data;
	dlist_t *cur;
	int cur_line = 1;

	/*
	** The top line will scroll off.
	*/
	line->line = -2;

	blist->top = blist->top->next;
	if (blist->bot->next != NULL)
		blist->bot = blist->bot->next;

	cur = blist->top;
	while (cur != NULL) {
		line = cur->data;

		line->line = cur_line++;
		if (cur == blist->bot)
			break;
		cur = cur->next;
	}

	line = blist->bot->data;

	scrollok(blist->win, TRUE);
	wscrl(blist->win, 1);
	scrollok(blist->win, FALSE);

	if (line->line != BLIST_LAST_ROW(blist)) {
		wmove(blist->win, BLIST_LAST_ROW(blist) + 1, 0);
		wclrtoeol(blist->win);
	}

	wmove(blist->win, line->line, 1);
	wclrtoeol(blist->win);
	mvwaddstr(blist->win, line->line, 2, line->text);

	blist_draw_border(blist);
}

struct blist_line *blist_get_cursor(struct buddy_list *blist) {
	if (blist->cursor != NULL)
		return (blist->cursor->data);

	return (NULL);
}

void blist_scroll_up(struct buddy_list *blist) {
	dlist_t *cur;
	int cur_line = 1;
	struct blist_line *line = blist->bot->data;

	blist->top = blist->top->prev;

	/*
	** The bottom row will scroll off if it's on the last line.
	*/

	if (line->line == BLIST_LAST_ROW(blist)) {
		line->line = -1;
		blist->bot = blist->bot->prev;
	}

	cur = blist->top;
	while (cur != NULL) {
		line = cur->data;

		line->line = cur_line++;
		if (cur == blist->bot)
			break;
		cur = cur->next;
	}

	line = blist->top->data;

	scrollok(blist->win, TRUE);
	wscrl(blist->win, -1);
	scrollok(blist->win, FALSE);

	wmove(blist->win, line->line, 1);
	wclrtoeol(blist->win);
	mvwaddstr(blist->win, line->line, 2, line->text);

	blist_draw_border(blist);
}

void blist_resize(struct buddy_list *blist, u_int32_t rows, u_int32_t cols) {
	dlist_t *cur;
	int line_num = 1;
	struct blist_line *cursor_line;

	/*
	** If the buddy list hasn't been initialized yet, there's nothing to do.
	*/

	if (blist->win == NULL)
		return;

	blist->rows = rows - STATUS_ROWS;

	wresize(blist->win, blist->rows, blist->cols);
	mvwin(blist->win, 0, cols - blist->cols);

	for (cur = blist->top ; cur != NULL ; cur = cur->next) {
		struct blist_line *line = cur->data;

		line->line = line_num++;
		blist->bot = cur;

		if (line_num > BLIST_LAST_ROW(blist)) {
			cur = cur->next;

			while (cur != NULL) {
				line = cur->data;

				line->line = -1;
				cur = cur->next;
			}

			break;
		}
	}

	cursor_line = blist->cursor->data;
	if (cursor_line->line < 0)
		blist->cursor = blist->top;

	blist->dirty = 1;
	blist_redraw(blist);
	blist_draw_curs(blist, 1);
}

static void blist_destroy_cleanup(void *data) {
	struct blist_line *line = data;

	free(line->text);
	free(line);
}

void blist_destroy(struct buddy_list *blist) {
	dlist_t *cur = screen.window_list;

	do {
		struct imwindow *imwindow = cur->data;

		if (imwindow->blist_visible)
			blist_toggle(imwindow->priv->blist, imwindow);

		cur = cur->next;
	} while (cur != screen.window_list);

	delwin(blist->win);
	dlist_destroy(blist->lines, blist_destroy_cleanup);
	memset(blist, 0, sizeof(*blist));
}

void blist_redraw(struct buddy_list *blist) {
	dlist_t *cur;

	wclear(blist->win);
	blist_draw_border(blist);

	for (cur = blist->top ; cur != NULL ; cur = cur->next) {
		struct blist_line *blist_line = cur->data;

		if (blist_line->line < 0)
			break;

		mvwaddstr(blist->win, blist_line->line, 2, blist_line->text);

		if (blist_line->type == TYPE_BUDDY) {
			struct buddy *buddy = blist_line->data;
			chtype indicator;

			indicator = blist_indicator(buddy->status);
			mvwaddch(blist->win, blist_line->line, 3, indicator);
		}
	}
}

void blist_refresh(struct buddy_list *blist, struct imwindow *imwindow) {
	if (!blist->dirty)
		return;

	if (imwindow->blist_visible) {
		wnoutrefresh(blist->win);
		blist->dirty = 0;
	}
}

#if 0
void debug(struct buddy_list *blist) {
	dlist_t *cur = blist->lines;
	struct blist_line *line;
	screen_status_msg("*************************************************");

	line = blist->bot->data;
	if (line->type == TYPE_GROUP) {
		struct group *gr = line->data;

		screen_status_msg("bottom: (%d) %s", line->line, gr->name);
	} else {
		struct buddy *buddy = line->data;

		screen_status_msg("bottom: (%d) %s", line->line, buddy->name);
	}

	line = blist->top->data;
	if (line->type == TYPE_GROUP) {
		struct group *gr = line->data;

		screen_status_msg("top: (%d) %s", line->line, gr->name);
	} else {
		struct buddy *buddy = line->data;

		screen_status_msg("top: (%d) %s", line->line, buddy->name);
	}
	line = blist->cursor->data;
	if (line->type == TYPE_GROUP) {
		struct group *gr = line->data;

		screen_status_msg("cursor: (%d) %s", line->line, gr->name);
	} else {
		struct buddy *buddy = line->data;

		screen_status_msg("cursor: (%d) %s", line->line, buddy->name);
	}

	while (cur != NULL) {
		line = cur->data;

		if (line->type == TYPE_GROUP) {
			struct group *gr = line->data;

			screen_status_msg("G: %s <=> %d", gr->name, line->line);
		} else {
			struct buddy *buddy = line->data;

			screen_status_msg("B: %s <=> %d", buddy->name, line->line);
		}

		cur = cur->next;
	}
}
#endif
