/*
** pork_command.c - interface to commands typed by the user
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <pork.h>
#include <pork_command_defs.h>

/*
** Note that the "struct command" arrays are arranged in alphabetical
** order.  They have to be like that.
*/

/*
** The / command set.
*/

static const struct command command[] = {
	{ "",			cmd_send			},
	{ "alias",		cmd_alias			},
	{ "away",		cmd_away			},
	{ "bind",		cmd_bind			},
	{ "blist",		cmd_blist			},
	{ "buddy",		cmd_buddy			},
	{ "connect",	cmd_connect			},
	{ "disconnect", cmd_disconnect		},
	{ "echo",		cmd_echo			},
	{ "email",		cmd_email			},
	{ "help",		cmd_help			},
	{ "history",	cmd_history			},
	{ "idle",		cmd_idle			},
	{ "input",		cmd_input			},
	{ "invite",		cmd_invite			},
	{ "join",		cmd_join			},
	{ "msg",		cmd_msg				},
	{ "part",		cmd_part			},
	{ "passwd",		cmd_passwd			},
	{ "profile",	cmd_profile,		},
	{ "query",		cmd_query			},
	{ "quit",		cmd_quit			},
	{ "refresh",	cmd_refresh			},
	{ "scroll",		cmd_scroll			},
	{ "set",		cmd_set				},
	{ "unalias",	cmd_unalias			},
	{ "unbind",		cmd_unbind			},
	{ "win",		cmd_window			},
};

/*
** /input commands
*/

static const struct command input_command[] = {
	{ "backspace",				cmd_input_bkspace			},
	{ "clear",					cmd_input_clear				},
	{ "clear_next_word",		cmd_input_clear_next		},
	{ "clear_prev_word",		cmd_input_clear_prev		},
	{ "clear_to_end",			cmd_input_clear_to_end		},
	{ "clear_to_start",			cmd_input_clear_to_start	},
	{ "delete",					cmd_input_delete			},
	{ "end",					cmd_input_end				},
	{ "focus_next",				cmd_input_focus_next		},
	{ "insert",					cmd_input_insert			},
	{ "left",					cmd_input_left				},
	{ "next_word",				cmd_input_next_word			},
	{ "prev_word",				cmd_input_prev_word			},
	{ "prompt",					cmd_input_prompt			},
	{ "right",					cmd_input_right				},
	{ "send",					cmd_input_send				},
	{ "start",					cmd_input_start				},
};

USER_COMMAND(cmd_input_bkspace) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_bkspace(imwindow->input);
}

USER_COMMAND(cmd_input_clear) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_clear_line(imwindow->input);
}

USER_COMMAND(cmd_input_clear_prev) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_clear_prev_word(imwindow->input);
}

USER_COMMAND(cmd_input_clear_next) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_clear_next_word(imwindow->input);
}

USER_COMMAND(cmd_input_clear_to_end) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_clear_to_end(imwindow->input);
}

USER_COMMAND(cmd_input_focus_next) {
	struct imwindow *imwindow = screen.cur_window->data;

	imwindow_switch_focus(imwindow);
	if (imwindow->blist_visible) {
		blist_draw_border(imwindow->priv->blist);
		wnoutrefresh(imwindow->priv->blist->win);
	}
}

USER_COMMAND(cmd_input_clear_to_start) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_clear_to_start(imwindow->input);
}

USER_COMMAND(cmd_input_delete) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_delete(imwindow->input);
}

USER_COMMAND(cmd_input_end) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_end(imwindow->input);
}

USER_COMMAND(cmd_input_insert) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args != NULL)
		input_insert_str(imwindow->input, args);
}

USER_COMMAND(cmd_input_left) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_move_left(imwindow->input);
}

USER_COMMAND(cmd_input_prev_word) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_prev_word(imwindow->input);
}

USER_COMMAND(cmd_input_prompt) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL) {
		input_set_prompt(imwindow->input, NULL);
		screen_status_msg("Input prompt cleared");
	} else {
		if (input_set_prompt(imwindow->input, args) == -1)
			screen_status_msg("Your prompt is too long for the screen");
		else {
			screen_status_msg("Input prompt set to %s",
				imwindow->input->prompt);
		}
	}
}

USER_COMMAND(cmd_input_next_word) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_next_word(imwindow->input);
}

USER_COMMAND(cmd_input_right) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_move_right(imwindow->input);
}

USER_COMMAND(cmd_input_send) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct input *input = imwindow->input;
	static int recursion;

	/*
	** This is kind of a hack, but it's necessary if the client
	** isn't to crash when someone types "/input send"
	*/

	if (recursion == 1 && args == NULL)
		return;

	recursion = 1;

	if (args != NULL) {
		xstrncpy(input->input_buf, args, sizeof(input->input_buf));
		input->len = strlen(input->input_buf);
	}

	if (input->len > 0) {
		char *input_str = xstrdup(input->input_buf);

		input_history_add(input);
		input_clear_line(input);

		if (*input_str == opt_get_char(OPT_CMDCHARS))
			run_command(&input_str[1], CMDSET_MAIN);
		else {
			if (!imwindow->priv->connected)
				screen_status_msg("You're not logged in");
			else if (imwindow->type == TYPE_IM) {
				chtype outbuf[INPUT_BUFFER_LEN + 128];

				aim_send_msg(imwindow->priv, imwindow->target, input_str);

				imwindow_prepare_input(imwindow, input_str,
					outbuf, sizeof(outbuf));
				swindow_input(&imwindow->swindow, outbuf);
				wnoutrefresh(imwindow->swindow.win);
			}
		}

		free(input_str);
	}

	recursion = 0;
}

USER_COMMAND(cmd_input_start) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_home(imwindow->input);
}

/*
** /win commands
*/

static const struct command scroll_command[] = {
	{ "by",						cmd_scroll_by				},
	{ "down",					cmd_scroll_down				},
	{ "end",					cmd_scroll_end				},
	{ "page_down",				cmd_scroll_pgdown			},
	{ "page_up",				cmd_scroll_pgup				},
	{ "start",					cmd_scroll_start			},
	{ "up",						cmd_scroll_up				},
};

USER_COMMAND(cmd_scroll_by) {
	int lines;
	char *end;

	lines = strtol(args, &end, 0);
	if (*end != '\0') {
		screen_status_msg("Bad number of lines: %s", args);
		return;
	}

	screen_scroll_by(&screen, lines);
}

USER_COMMAND(cmd_scroll_down) {
	screen_scroll_down(&screen);
}

USER_COMMAND(cmd_scroll_end) {
	screen_scroll_to_end(&screen);
}

USER_COMMAND(cmd_scroll_pgdown) {
	screen_scroll_pgdown(&screen);
}

USER_COMMAND(cmd_scroll_pgup) {
	screen_scroll_pgup(&screen);
}

USER_COMMAND(cmd_scroll_start) {
	screen_scroll_to_start(&screen);
}

USER_COMMAND(cmd_scroll_up) {
	screen_scroll_up(&screen);
}

/*
** window commands.
*/

static const struct command window_command[] = {
	{ "bind",					cmd_window_bind				},
	{ "clear",					cmd_window_clear			},
	{ "close",					cmd_window_close			},
	{ "erase",					cmd_window_erase			},
	{ "list",					cmd_window_list				},
	{ "next",					cmd_window_next				},
	{ "num",					cmd_window_num				},
	{ "prev",					cmd_window_prev				},
	{ "rename",					cmd_window_rename			},
	{ "set",					cmd_window_set				},
	{ "swap",					cmd_window_swap				},
};

USER_COMMAND(cmd_window_bind) {
	struct imwindow *imwindow = screen.cur_window->data;
	int ret;

	if (args == NULL || blank_str(args)) {
		if (imwindow->priv != NULL && imwindow->priv->username != NULL) {
			screen_status_msg("This window is bound to account: %s",
				imwindow->priv->username);
		} else
			screen_status_msg("This window is bound to no account");

		return;
	}

	ret = imwindow_bind_acct(imwindow, args);
	if (ret == -1)
		screen_status_msg("Account %s isn't signed on", args);
	else {
		screen_status_msg("This window is now bound to account: %s",
			imwindow->priv->username);
	}
}

USER_COMMAND(cmd_window_clear) {
	screen_clear(&screen);
}

USER_COMMAND(cmd_window_close) {
	screen_close_window(&screen);
}

USER_COMMAND(cmd_window_erase) {
	screen_erase(&screen);
}

USER_COMMAND(cmd_window_list) {
	char buf[4096];
	size_t i = 0;
	size_t len = sizeof(buf);
	int ret;
	dlist_t *cur;
	static const char *win_types[] = { "Status", "IM", "Chat Room" };

	screen_status_msg("Window List:");
	cur = screen.window_list;
	do {
		struct imwindow *imwindow = cur->data;

		ret = snprintf(&buf[i], len, "Window refnum %u: Name=%s, Type=%s, Target=%s\n",
				imwindow->refnum, imwindow->name,
				win_types[imwindow->type], imwindow->target);
		i += ret;
		len -= ret;

		cur = cur->next;
	} while (cur != screen.window_list);

	screen_print_multiline(screen.cur_window->data, buf, 0);
	imwindow_refresh(screen.cur_window->data);
}

USER_COMMAND(cmd_window_next) {
	screen_cycle_fwd(&screen);
}

USER_COMMAND(cmd_window_num) {
	u_int32_t num;
	char *end;
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL) {
		screen_status_msg("This is window %u", imwindow->refnum);
		return;
	}

	num = strtoul(args, &end, 0);
	if (*end != '\0') {
		screen_status_msg("Bad window number: %s", args);
	}

	screen_renumber(&screen, num);
	screen_status_msg("This is now window %u", imwindow->refnum);
}

USER_COMMAND(cmd_window_prev) {
	screen_cycle_bak(&screen);
}

USER_COMMAND(cmd_window_rename) {
	struct imwindow *imwindow = screen.cur_window->data;

	imwindow_rename(imwindow, xstrdup(args));
}

USER_COMMAND(cmd_window_set) {
	char *variable;
	char *value;
	int opt;
	struct imwindow *imwindow = screen.cur_window->data;

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		wopt_print(imwindow);
		return;
	}

	opt = wopt_find(variable);
	if (opt == -1) {
		screen_status_msg("Unknown variable: %s", variable);
		return;
	}

	value = strsep(&args, "\0");
	if (value == NULL || blank_str(value)) {
		wopt_print_var(imwindow, opt);
		return;
	}

	if (wopt_set(imwindow, opt, value) == -1)
		screen_status_msg("Bad argument for %s: %s", variable, value);
	else
		screen_status_msg("%s set to %s", variable, value);
}

/*
** History Manipulation.
*/

static const struct command history_command[] = {
	{ "clear",					cmd_history_clear			},
	{ "next",					cmd_history_next			},
	{ "prev",					cmd_history_prev			},
};

USER_COMMAND(cmd_history_clear) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_history_clear(imwindow->input);
}

USER_COMMAND(cmd_history_next) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_history_next(imwindow->input);
}

USER_COMMAND(cmd_history_prev) {
	struct imwindow *imwindow = screen.cur_window->data;

	input_history_prev(imwindow->input);
}

/*
** /buddy commands
*/

static const struct command buddy_command[] = {
	{ "add",			cmd_buddy_add			},
	{ "add_block",		cmd_buddy_add_block		},
	{ "add_group",		cmd_buddy_add_group		},
	{ "add_permit",		cmd_buddy_add_permit	},
	{ "alias",			cmd_buddy_alias			},
	{ "away",			cmd_buddy_away			},
	{ "print_block",	cmd_buddy_print_block	},
	{ "print_list",		cmd_buddy_print_list	},
	{ "print_permit",	cmd_buddy_print_permit	},
	{ "profile",		cmd_buddy_profile		},
	{ "remove",			cmd_buddy_remove		},
	{ "remove_block",	cmd_buddy_remove_block	},
	{ "remove_group",	cmd_buddy_remove_group	},
	{ "remove_permit",	cmd_buddy_remove_permit	},
	{ "warn",			cmd_buddy_warn			},
	{ "warn_anon",		cmd_buddy_warn_anon		},
};

USER_COMMAND(cmd_buddy_add) {
	char *screen_name;
	char *group_name;
	struct group *group;
	struct buddy *buddy;
	struct imwindow *imwindow = screen.cur_window->data;

	screen_name = strsep(&args, " ");
	if (screen_name == NULL) {
		screen_status_msg("syntax is /buddy add <user> <group>");
		return;
	}

	buddy = buddy_find(imwindow->priv, screen_name);
	if (buddy != NULL) {
		screen_status_msg("%s is already a member of the group %s",
			screen_name, buddy->group->name);
		return;
	}

	group_name = strsep(&args, " ");
	if (group_name == NULL) {
		screen_status_msg("syntax is /buddy add <user> <group>");
		return;
	}

	group = group_find(imwindow->priv, group_name);
	if (group == NULL) {
		screen_status_msg("The group %s does not exist", group_name);
		return;
	}

	buddy_add(imwindow->priv, screen_name, group);
}

USER_COMMAND(cmd_buddy_add_block) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_block(imwindow->priv, args) == -1)
		screen_status_msg("%s is already blocked", args);
	else
		screen_status_msg("%s added to block list", args);
}

USER_COMMAND(cmd_buddy_add_permit) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_permit(imwindow->priv, args) == -1)
		screen_status_msg("%s is already permitted", args);
	else
		screen_status_msg("%s added to permit list", args);
}

USER_COMMAND(cmd_buddy_add_group) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	group_add(imwindow->priv, args);
}

USER_COMMAND(cmd_buddy_alias) {
	struct buddy *buddy;
	char *buddy_name;
	char *alias;
	struct imwindow *imwindow = screen.cur_window->data;

	buddy_name = strsep(&args, " ");
	if (buddy_name == NULL)
		return;

	alias = strsep(&args, " ");
	if (alias == NULL || blank_str(alias))
		return;	

	buddy = buddy_find(imwindow->priv, buddy_name);
	if (buddy == NULL) {
		screen_status_msg("%s is not on your buddy list", buddy_name);
		return;
	}

	screen_status_msg("alias: %s -> %s", buddy_name,alias);
	buddy_alias(imwindow->priv, buddy, alias);
}

USER_COMMAND(cmd_buddy_away) {
	char nname[NUSER_LEN];
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args)) {
		aim_whois(imwindow->priv, imwindow->priv->username);
		return;
	}

	normalize(args, nname, sizeof(nname));
	aim_get_away_msg(imwindow->priv, nname);
}

USER_COMMAND(cmd_buddy_print_block) {
	struct imwindow *imwindow = screen.cur_window->data;

	buddy_print_block(imwindow->priv);
}

USER_COMMAND(cmd_buddy_print_list) {
	struct imwindow *imwindow = screen.cur_window->data;

	buddy_print_list(imwindow->priv);
}

USER_COMMAND(cmd_buddy_profile) {
	char nname[NUSER_LEN];
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args)) {
		aim_whois(imwindow->priv, imwindow->priv->username);
		return;
	}

	normalize(args, nname, sizeof(nname));
	aim_whois(imwindow->priv, nname);
}

USER_COMMAND(cmd_buddy_print_permit) {
	struct imwindow *imwindow = screen.cur_window->data;

	buddy_print_permit(imwindow->priv);
}

USER_COMMAND(cmd_buddy_remove_permit) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_permit(imwindow->priv, args) != 0)
		screen_status_msg("%s isn't on your permit list", args);
	else
		screen_status_msg("%s removed from permit list", args);
}

USER_COMMAND(cmd_buddy_remove_block) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_block(imwindow->priv, args) != 0)
		screen_status_msg("%s isn't on your block list", args);
	else
		screen_status_msg("%s removed from block list", args);
}

USER_COMMAND(cmd_buddy_remove) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove(imwindow->priv, args) != 0)
		screen_status_msg("%s isn't on your buddy list", args);
	else
		screen_status_msg("%s removed from buddy list", args);
}

USER_COMMAND(cmd_buddy_remove_group) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL || blank_str(args))
		return;

	if (group_remove(imwindow->priv, args) != 0)
		screen_status_msg("No such group: %s", args);
	else
		screen_status_msg("Removed group: %s", args);
}

USER_COMMAND(cmd_buddy_warn) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args))
		return;

	aim_warn(imwindow->priv, args);
}

USER_COMMAND(cmd_buddy_warn_anon) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args))
		return;

	aim_warn_anon(imwindow->priv, args);
}

/*
** /blist commands
*/

static const struct command blist_command[] = {
	{ "add_block",		cmd_blist_add_block		},
	{ "add_permit",		cmd_blist_add_permit	},
	{ "away",			cmd_blist_away			},
	{ "collapse",		cmd_blist_collapse		},
	{ "down",			cmd_blist_down			},
	{ "end",			cmd_blist_end			},
	{ "goto",			cmd_blist_goto			},
	{ "idle",			cmd_blist_idle			},
	{ "invite",			cmd_blist_invite		},
	{ "move",			cmd_blist_move			},
	{ "page_down",		cmd_blist_pgdown		},
	{ "page_up",		cmd_blist_pgup			},
	{ "profile",		cmd_blist_profile		},
	{ "refresh",		cmd_blist_refresh		},
	{ "remove",			cmd_blist_remove		},
	{ "remove_block",	cmd_blist_remove_block	},
	{ "remove_permit",	cmd_blist_remove_permit	},
	{ "select",			cmd_blist_select		},
	{ "show",			cmd_blist_show			},
	{ "top",			cmd_blist_top			},
	{ "up",				cmd_blist_up			},
	{ "warn",			cmd_blist_warn			},
	{ "warn_anon",		cmd_blist_warn_anon		},
};

USER_COMMAND(cmd_blist_add_block) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_add_block(buddy->nname);
}

USER_COMMAND(cmd_blist_add_permit) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_add_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_away) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_away(buddy->nname);
}

USER_COMMAND(cmd_blist_collapse) {
	struct blist_line *line;
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	if (args == NULL) {
		line = blist_get_cursor(priv->blist);
		if (line == NULL || line->type != TYPE_GROUP)
			return;
	} else {
		struct group *gr = group_find(imwindow->priv, args);
		if (gr == NULL)
			return;

			line = gr->blist_line;
	}

	blist_collapse_group(priv->blist, line);
	blist_refresh(priv->blist, screen.cur_window->data);
}

USER_COMMAND(cmd_blist_down) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	blist_cursor_down(priv->blist);
	blist_refresh(priv->blist, screen.cur_window->data);
}

USER_COMMAND(cmd_blist_end) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	(void) priv;
}

USER_COMMAND(cmd_blist_goto) {
	if (args != NULL)
		cmd_query(args);
}

USER_COMMAND(cmd_blist_idle) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	blist_show_idle(priv->blist);
}

USER_COMMAND(cmd_blist_invite) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;
}

USER_COMMAND(cmd_blist_move) {
}

USER_COMMAND(cmd_blist_pgdown) {
}

USER_COMMAND(cmd_blist_pgup) {
}

USER_COMMAND(cmd_blist_profile) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_profile(buddy->nname);
}

USER_COMMAND(cmd_blist_refresh) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	redrawwin(priv->blist->win);
	blist_refresh(priv->blist, screen.cur_window->data);
}

USER_COMMAND(cmd_blist_remove) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_remove(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_block) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_remove_block(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_permit) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_remove_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_select) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line;

	line = blist_get_cursor(priv->blist);
	if (line == NULL)
		return;

	if (line->type == TYPE_GROUP)
		cmd_blist_collapse(NULL);
	else {
		struct buddy *buddy = line->data;

		cmd_blist_goto(buddy->nname);
	}
}

USER_COMMAND(cmd_blist_show) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	blist_toggle(priv->blist, screen.cur_window->data);
}

USER_COMMAND(cmd_blist_top) {
}

USER_COMMAND(cmd_blist_up) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;

	blist_cursor_up(priv->blist);
	blist_refresh(priv->blist, screen.cur_window->data);
}

USER_COMMAND(cmd_blist_warn) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_warn(buddy->nname);
}

USER_COMMAND(cmd_blist_warn_anon) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	struct blist_line *line = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (line == NULL || line->type == TYPE_GROUP)
		return;
	buddy = line->data;

	cmd_buddy_warn_anon(buddy->nname);
}

static const struct {
	const struct command *set;
	size_t elem;
	char *type;
} command_set[] = {
	{ command,			array_elem(command),			"" 			},
	{ window_command,	array_elem(window_command),		"window "	},
	{ history_command,	array_elem(history_command),	"history "	},
	{ input_command,	array_elem(input_command),		"input "	},
	{ scroll_command,	array_elem(scroll_command),		"scroll "	},
	{ buddy_command,	array_elem(buddy_command),		"buddy "	},
	{ blist_command,	array_elem(blist_command),		"blist "	},
};

/*
** Main command set.
*/

USER_COMMAND(cmd_alias) {
	char *alias;
	char *str;

	alias = strsep(&args, " ");
	if (alias == NULL || blank_str(alias)) {
		hash_iterate(&screen.alias_hash, print_alias, NULL);
		return;
	}

	str = strsep(&args, "\0");
	if (str == NULL || blank_str(str)) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_status_msg("%s is aliased to %s",
				lalias->alias, lalias->str);
		} else
			screen_status_msg("There is no alias for %s", alias);

		return;
	}

	if (alias_add(&screen.alias_hash, alias, str) == 0) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_status_msg("%s is aliased to %s",
				lalias->alias, lalias->str);
			return;
		}
	}

	screen_status_msg("Error adding alias for %s", alias);
}

USER_COMMAND(cmd_away) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}

	aim_set_away(imwindow->priv, args);
}

USER_COMMAND(cmd_bind) {
	int key;
	char *func;
	char *key_str;
	struct imwindow *imwindow = screen.cur_window->data;
	struct key_binds *target_binds = imwindow->active_binds;
	struct binding *binding;

	key_str = strsep(&args, " ");
	if (key_str == NULL || blank_str(key_str)) {
		hash_iterate(&target_binds->hash, print_binding, NULL);
		return;
	}

	if (key_str[0] == '-' && key_str[1] != '\0') {
		if (!strcasecmp(key_str, "-b") || !strcasecmp(key_str, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(key_str, "-m") || !strcasecmp(key_str, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_status_msg("Bad bind flag: %s", key_str);
			return;
		}

		key_str = strsep(&args, " ");

		if (key_str == NULL || blank_str(key_str)) {
			hash_iterate(&target_binds->hash, print_binding, NULL);
			return;
		}
	}

	key = bind_get_keycode(key_str);
	if (key == -1) {
		screen_status_msg("Bad keycode: %s", key_str);
		return;
	}

	func = strsep(&args, "\0");
	if (func != NULL) {
		if (*func == opt_get_char(OPT_CMDCHARS) && *(func + 1) != '\0')
			func++;
		if (blank_str(func))
			func = NULL;
	}

	if (func == NULL) {
		binding = bind_find(target_binds, key);
		if (binding != NULL)
			screen_status_msg("%s is bound to %s", key_str, binding->binding);
		else
			screen_status_msg("%s is not bound", key_str);

		return;
	}

	bind_add(target_binds, key, func);
	binding = bind_find(target_binds, key);
	if (binding != NULL) {
		screen_status_msg("%s is bound to %s", key_str, binding->binding);
		return;
	}

	screen_status_msg("Error binding %s", key_str);
}

USER_COMMAND(cmd_connect) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv;

	if (args != NULL && !blank_str(args)) {
		struct aim_acct *acct;

		acct = screen_add_acct(&screen, args);
		if (acct == NULL) {
			screen_status_msg("You're already connected as %s",
				imwindow->priv->username);
			return;
		}

		priv = acct->priv;
		if (imwindow->priv == &screen.null_priv) {
			cmd_window_bind(acct->username);

			if (!imwindow->blist_visible) {
				if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BUDDY_LIST))
					wopt_set(imwindow, WOPT_SHOW_BUDDY_LIST, "1");
			}

		}
	} else {
		if (imwindow->priv->connected) {
			cmd_window_bind(imwindow->priv->username);
			return;
		}

		priv = imwindow->priv;
	}

	if (priv->passwd == NULL) {
		char buf[128];

		screen_get_passwd(buf, sizeof(buf));
		if (buf[0] == '\0') {
			/* EH */
			screen_status_msg("There was an error reading your password");
			return;
		}

		priv->passwd = xstrdup(buf);
		memset(buf, 0, sizeof(buf));
	}

	if (aim_login(priv) == -1)
		screen_status_msg("Unable to login as %s", priv->username);
}

USER_COMMAND(cmd_echo) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args != NULL) {
		chtype *ch;
		size_t len = strlen(args) + 1;

		ch = xmalloc(sizeof(chtype) * len);
		plaintext_to_cstr(args, ch, len);
		swindow_add(&imwindow->swindow, ch, 0);
		imwindow_refresh(imwindow);
	}
		
}

USER_COMMAND(cmd_disconnect) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args != NULL && !blank_str(args)) {
		if (screen_del_acct(&screen, args) != 0)
			screen_status_msg("You're not logged in as %s", args);
	} else {
		if (screen_del_acct(&screen, imwindow->priv->username) != 0) {
			screen_status_msg("Account %s is not logged in",
				imwindow->priv->username);
		} 
	}
}

USER_COMMAND(cmd_help) {
}

USER_COMMAND(cmd_idle) {
	char *end;
	u_int32_t idle_secs;
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}

	if (args == NULL) {
		aim_set_idle(imwindow->priv, 0);
		return;
	}

	idle_secs = strtoul(args, &end, 10);
	if (*end != '\0') {
		screen_status_msg("Bad argument: %s", args);
		return;
	}

	aim_set_idle(imwindow->priv, idle_secs);
}

USER_COMMAND(cmd_invite) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_join) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_msg) {
	struct imwindow *conv_window;
	char *target;
	char *msg;
	chtype outbuf[INPUT_BUFFER_LEN + 128];
	struct imwindow *imwindow = screen.cur_window->data;

	if (args == NULL)
		return;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not logged in");
		return;
	}

	target = strsep(&args, " ");
	if (target == NULL)
		return;
	msg = strsep(&args, "\0");
	if (msg == NULL)
		return;

	conv_window = imwindow_find(target);
	if (conv_window == NULL) {
		conv_window = screen_new_window(&screen, imwindow->priv, target);
		if (conv_window == NULL)
			return;
		screen_goto_window(&screen, conv_window->refnum);
	}

	aim_send_msg(conv_window->priv, target, msg);

	imwindow_prepare_input(imwindow, msg, outbuf, sizeof(outbuf));
	swindow_input(&conv_window->swindow, outbuf);

	if (screen.cur_window->data == conv_window)
		wnoutrefresh(conv_window->swindow.win);
}

USER_COMMAND(cmd_part) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_query) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (args != NULL && !blank_str(args)) {
		struct imwindow *conv_window;

		conv_window = imwindow_find(args);
		if (conv_window == NULL)
			conv_window = screen_new_window(&screen, imwindow->priv, args);

		screen_goto_window(&screen, conv_window->refnum);
	} else
		screen_close_window(&screen);
}

USER_COMMAND(cmd_quit) {
	pork_signoff();
}

USER_COMMAND(cmd_refresh) {
	screen_refresh(&screen);
}

USER_COMMAND(cmd_send) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}

	if (imwindow->type == TYPE_IM && args != NULL)
		aim_send_msg(imwindow->priv, imwindow->target, args);
}

USER_COMMAND(cmd_unbind) {
	char *binding;
	struct imwindow *imwindow = screen.cur_window->data;
	struct key_binds *target_binds = imwindow->active_binds;
	int c;

	binding = strsep(&args, " ");
	if (binding == NULL || blank_str(binding))
		return;

	if (binding[0] == '-' && binding[1] != '\0') {
		if (!strcasecmp(binding, "-b") || !strcasecmp(binding, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(binding, "-m") || !strcasecmp(binding, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_status_msg("Bad unbind flag: %s", binding);
			return;
		}

		binding = strsep(&args, " ");

		if (binding == NULL || blank_str(binding))
			return;
	}

	c = bind_get_keycode(binding);
	if (c == -1) {
		screen_status_msg("Bad keycode: %s", binding);
		return;
	}

	if (bind_remove(target_binds, c) == -1)
		screen_status_msg("There is no binding for %s", binding);
	else
		screen_status_msg("Binding for %s removed", binding);
}

USER_COMMAND(cmd_unalias) {
	char *alias = strsep(&args, " ");

	if (alias == NULL)
		return;

	if (alias_remove(&screen.alias_hash, alias) == -1)
		screen_status_msg("%s: No such alias", alias);
	else
		screen_status_msg("Alias %s removed", alias);
}

USER_COMMAND(cmd_email) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_passwd) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_profile) {
	struct imwindow *imwindow = screen.cur_window->data;

	aim_set_profile(imwindow->priv, args);
}

USER_COMMAND(cmd_window_swap) {
	u_int32_t num;
	char *end;

	if (args == NULL)
		return;

	num = strtoul(args, &end, 0);
	if (*end != '\0')
		return;

	screen_goto_window(&screen, num);
}

USER_COMMAND(cmd_window) {
	if (args != NULL)
		run_command(args, CMDSET_WIN);
}

USER_COMMAND(cmd_buddy) {
	if (args != NULL)
		run_command(args, CMDSET_BUDDY);
}

USER_COMMAND(cmd_blist) {
	if (args != NULL)
		run_command(args, CMDSET_BLIST);
}

USER_COMMAND(cmd_input) {
	if (args != NULL)
		run_command(args, CMDSET_INPUT);
}

USER_COMMAND(cmd_history) {
	if (args != NULL)
		run_command(args, CMDSET_HISTORY);
}

USER_COMMAND(cmd_scroll) {
	if (args != NULL)
		run_command(args, CMDSET_SCROLL);
}

USER_COMMAND(cmd_set) {
	char *variable;
	char *value;
	int opt;

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		opt_print();
		return;
	}

	opt = opt_find(variable);
	if (opt == -1) {
		screen_status_msg("Unknown variable: %s", variable);
		return;
	}

	value = strsep(&args, "\0");
	if (value == NULL || blank_str(value)) {
		opt_print_var(opt);
		return;
	}

	if (opt_set(opt, value) == -1) {
		screen_status_msg("Bad argument for %s: %s", variable, value);
	} else
		screen_status_msg("%s set to %s", variable, value);
}

int run_command(char *str, u_int32_t set) {
	static char buf[INPUT_BUFFER_LEN * 4];
	struct alias *alias;
	char *cmd_str;
	struct command *cmd;

	xstrncpy(buf, str, sizeof(buf));
	str = buf;

	cmd_str = strsep(&str, " ");
	if (set == CMDSET_MAIN) {
		alias = alias_find(&screen.alias_hash, cmd_str);

		if (alias != NULL) {
			char *p;
			char *alias_cmd;
			char *alias_args;
			static char alias_buf[INPUT_BUFFER_LEN];

			xstrncpy(alias_buf, alias->str, sizeof(alias_buf));

			p = alias_buf;
			alias_cmd = strsep(&p, " ");
			alias_args = strsep(&p, "\0");

			if (alias_args != NULL) {
				if (str == NULL) {
					str = alias_args;
					cmd_str = alias_cmd;
				} else {
					size_t aarg_len = strlen(alias_args);
					size_t str_len = strlen(str);

					if (aarg_len + str_len + 1 < sizeof(buf) - str_len) {
						memmove(&str[aarg_len + 1], str, str_len + 1);
						str[aarg_len] = ' ';
						memcpy(str, alias_args, aarg_len);
						cmd_str = alias_cmd;
					} else
						screen_status_msg("The alias is too long");
				}
			} else
				cmd_str = alias_cmd;
		}
	}

	cmd = bsearch(cmd_str, command_set[set].set,
			command_set[set].elem, sizeof(struct command), cmd_compare);

	if (cmd == NULL) {
		screen_status_msg("Unknown %scommand: %s",
			command_set[set].type, cmd_str);

		return (-1);
	}

	cmd->cmd(str);
	return (0);
}

static void print_binding(void *data, void *nothing __notused) {
	struct binding *binding = data;
	char key_name[32];

	bind_get_keyname(binding->key, key_name, sizeof(key_name));
	screen_status_msg("%s is bound to %s", key_name, binding->binding);
}

static void print_alias(void *data, void *nothing __notused) {
	struct alias *alias = data;

	screen_status_msg("%s is aliased to %s", alias->alias, alias->str);
}

static int cmd_compare(const void *l, const void *r) {
	char *key = (char *) l;
	struct command *cmd = (struct command *) r;

	return (strcasecmp(key, cmd->name));
}
