/*
** pork_set.c - /SET command implementation.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>

#include <pork.h>

static int opt_set_bool(u_int32_t opt, char *args);
static int opt_set_char(u_int32_t opt, char *args);
static int opt_set_int(u_int32_t opt, char *args);
static int opt_set_str(u_int32_t opt, char *args);

static int wopt_set_bool(struct imwindow *imwindow, u_int32_t opt, char *args);
static int wopt_set_char(struct imwindow *imwindow, u_int32_t opt, char *args);
static int wopt_set_int(struct imwindow *imwindow, u_int32_t opt, char *args);
static int wopt_set_str(struct imwindow *imwindow, u_int32_t opt, char *args);

static void wopt_changed_histlen(struct imwindow *imwindow);
static void wopt_changed_log(struct imwindow *imwindow);
static void wopt_changed_logfile(struct imwindow *imwindow);
static void wopt_changed_priv_input(struct imwindow *imwindow);
static void wopt_changed_scrollbuf_len(struct imwindow *imwindow);
static void wopt_changed_scroll_on_out(struct imwindow *imwindow);
static void wopt_changed_scroll_on_in(struct imwindow *imwindow);
static void wopt_changed_show_blist(struct imwindow *imwindow);
static void wopt_changed_timestamp(struct imwindow *imwindow);

/*
** Name:		The name of the option
** Type:		The type of the option (boolean, string, int, char)
** Dynamic:		When the option changes, does the current value need to
**				be freed before the new one is set (i.e. was it
**				dynamically allocated?).
** Set func:	The function used to set the value of the option.
** Change func:	The function to be called when the value of the option changes.
** Default:		The default value of the option.
*/

/*
**	NAME					TYPE	DYNAMIC		SET FUNC	CHANGE FUNC		DEFAULT VAL
*/
static struct global_var global_var[] = {
	{ "AUTO_RECONNECT",		OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
	{ "BANNER",				OPT_STR,	0,	opt_set_str,	NULL,	SET_STR("*** ")		},
	{ "BEEP_ON_OUTPUT",		OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "CMDCHARS",			OPT_CHAR,	0,	opt_set_char,	NULL,	SET_CHAR('/')		},
	{ "HISTORY_LEN",		OPT_INT,	0,	opt_set_int,	NULL,	SET_INT(400)		},
	{ "LOG",				OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "LOGIN_ON_STARTUP",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
	{ "PORK_DIR",			OPT_STR,	0,	opt_set_str,	NULL,	SET_STR(NULL)		},
	{ "PRIVATE_INPUT",		OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "REPORT_IDLE",		OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
	{ "SCROLLBUF_LEN",		OPT_INT,	0,	opt_set_int, 	NULL,	SET_INT(0xffffffff)	},
	{ "SCROLL_ON_INPUT",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
	{ "SCROLL_ON_OUTPUT",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "SEND_REMOVES_AWAY",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
	{ "SHOW_BUDDY_IDLE",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "SHOW_BUDDY_LIST",	OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(0)			},
	{ "TIMESTAMP",			OPT_BOOL,	0,	opt_set_bool,	NULL,	SET_BOOL(1)			},
};

/*
** Name:		The name of the window specific option.
** Type:		The type of the option (boolean, string, int, char)
** Set func:	The function used to set the option.
** Change func:	The function to be called when the option changes.
*/

/*
**		NAME				TYPE		SET FUNC		CHANGE FUNC
*/

static struct window_var window_var[] = {
	{ "BEEP_ON_OUTPUT",		OPT_BOOL,	wopt_set_bool,	NULL						},
	{ "HISTORY_LEN",		OPT_INT,	wopt_set_int,	wopt_changed_histlen		},
	{ "LOG",				OPT_BOOL,	wopt_set_bool,	wopt_changed_log			},
	{ "LOGFILE",			OPT_STR,	wopt_set_str, 	wopt_changed_logfile		},
	{ "PRIVATE_INPUT",		OPT_BOOL,	wopt_set_bool,	wopt_changed_priv_input		},
	{ "SCROLLBUF_LEN",		OPT_INT,	wopt_set_int,	wopt_changed_scrollbuf_len	},
	{ "SCROLL_ON_INPUT",	OPT_BOOL,	wopt_set_bool,	wopt_changed_scroll_on_in	},
	{ "SCROLL_ON_OUTPUT",	OPT_BOOL,	wopt_set_bool,	wopt_changed_scroll_on_out	},
	{ "SHOW_BUDDY_LIST",	OPT_BOOL,	wopt_set_bool,	wopt_changed_show_blist		},
	{ "TIMESTAMP",			OPT_BOOL,	wopt_set_bool,	wopt_changed_timestamp		},
};

/*
** Print out the value of the specified option.
*/

void wopt_print_var(struct imwindow *imwindow, int i) {
	switch (window_var[i].type) {
		case OPT_BOOL:
			screen_status_msg("%s is set to %s", window_var[i].name,
				(imwindow->opts[i].val.b ? "TRUE" : "FALSE"));
			break;

		case OPT_STR:
			if (imwindow->opts[i].val.s != NULL) {
				screen_status_msg("%s is set to %s", window_var[i].name,
					imwindow->opts[i].val.s);
			} else
				screen_status_msg("%s is <UNSET>", window_var[i].name);
			break;

		case OPT_INT:
			screen_status_msg("%s is set to %d", window_var[i].name,
				imwindow->opts[i].val.i);
			break;

		case OPT_CHAR:
			screen_status_msg("%s is set to %c", window_var[i].name,
				imwindow->opts[i].val.c);
			break;
	}
}

/*
** Print the values of all the window-specific options.
*/

void wopt_print(struct imwindow *imwindow) {
	size_t i;

	for (i = 0 ; i < array_elem(window_var) ; i++)
		wopt_print_var(imwindow, i);
}

/*
** Print value the specified global option.
*/

void opt_print_var(int i) {
	switch (global_var[i].type) {
		case OPT_BOOL:
			screen_status_msg("%s is set to %s", global_var[i].name,
				(global_var[i].val.b ? "TRUE" : "FALSE"));
			break;

		case OPT_STR:
			if (global_var[i].val.s != NULL) {
				screen_status_msg("%s is set to %s", global_var[i].name,
					global_var[i].val.s);
			} else
				screen_status_msg("%s is <UNSET>", global_var[i].name);

			break;

		case OPT_INT:
			screen_status_msg("%s is set to %d", global_var[i].name,
				global_var[i].val.i);
			break;

		case OPT_CHAR:
			screen_status_msg("%s is set to %c", global_var[i].name,
				global_var[i].val.c);
			break;
	}

}

/*
** Print the values of all the global variables.
*/

void opt_print(void) {
	size_t i;

	for (i = 0 ; i < array_elem(global_var) ; i++)
		opt_print_var(i);
}

static int opt_compare(const void *l, const void *r) {
	const char *str = l;
	const struct global_var *gvar = r;

	return (strcasecmp(str, gvar->name));
}

static int wopt_compare(const void *l, const void *r) {
	const char *str = l;
	const struct window_var *wvar = r;

	return (strcasecmp(str, wvar->name));
}

/*
** Find the position of the global option named "name"
** in the global option table.
*/

int opt_find(const char *name) {
	struct global_var *gvar;
	u_int32_t offset;

	gvar = bsearch(name, global_var, array_elem(global_var),
				sizeof(struct global_var), opt_compare);

	if (gvar == NULL)
		return (-1);

	offset = (u_int32_t) gvar - (u_int32_t) &global_var[0];
	return (offset / sizeof(struct global_var));
}

/*
** Same as above, only for window-specific options.
*/

int wopt_find(const char *name) {
	struct window_var *wvar;
	u_int32_t offset;

	wvar = bsearch(name, window_var, array_elem(window_var),
				sizeof(struct window_var), wopt_compare);

	if (wvar == NULL)
		return (-1);

	offset = (u_int32_t) wvar - (u_int32_t) &window_var[0];
	return (offset / sizeof(struct window_var));
}

/*
** Return the value of the global option specified
** by "opt" that's of the specified type.
*/

inline u_int32_t opt_get_int(u_int32_t opt) {
	return (global_var[opt].val.i);
}

inline u_int32_t opt_get_bool(u_int32_t opt) {
	return (global_var[opt].val.b);
}

inline char opt_get_char(u_int32_t opt) {
	return (global_var[opt].val.c);
}

inline char *opt_get_str(u_int32_t opt) {
	return (global_var[opt].val.s);
}

/*
** Same as above, only for window-specific options.
*/

inline u_int32_t wopt_get_int(struct window_opts *wopt, u_int32_t opt) {
	return (wopt[opt].val.i);
}

inline char *wopt_get_str(struct window_opts *wopt, u_int32_t opt) {
	return (wopt[opt].val.s);
}

inline char wopt_get_char(struct window_opts *wopt, u_int32_t opt) {
	return (wopt[opt].val.c);
}

inline u_int32_t wopt_get_bool(struct window_opts *wopt, u_int32_t opt) {
	return (wopt[opt].val.b);
}

static void wopt_changed_histlen(struct imwindow *imwindow) {
	u_int32_t new_len = wopt_get_int(imwindow->opts, WOPT_HISTORY_LEN);

	imwindow->input->history_len = new_len;
	input_history_prune(imwindow->input);
}

static void wopt_changed_log(struct imwindow *imwindow) {
	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_LOG);

	if (new_val != imwindow->swindow.logged) {
		if (new_val == 0)
			swindow_end_log(&imwindow->swindow);
		else
			swindow_set_log(&imwindow->swindow);
	}
}

static void wopt_changed_logfile(struct imwindow *imwindow) {
	swindow_set_logfile(&imwindow->swindow,
		wopt_get_str(imwindow->opts, WOPT_LOGFILE));
}

static void wopt_changed_priv_input(struct imwindow *imwindow) {
	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);

	(void) new_val;
	/* XXX - fix this */

}

static void wopt_changed_scrollbuf_len(struct imwindow *imwindow) {
	u_int32_t new_len = wopt_get_int(imwindow->opts, WOPT_SCROLLBUF_LEN);

	imwindow->swindow.scrollbuf_max = new_len;
	swindow_prune(&imwindow->swindow);
}

static void wopt_changed_scroll_on_out(struct imwindow *imwindow) {
	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_OUTPUT);

	imwindow->swindow.scroll_on_output = new_val;	
}

static void wopt_changed_scroll_on_in(struct imwindow *imwindow) {
	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SCROLL_ON_INPUT);

	imwindow->swindow.scroll_on_input = new_val;	
}

static void wopt_changed_show_blist(struct imwindow *imwindow) {
	u_int32_t new_val = wopt_get_bool(imwindow->opts, WOPT_SHOW_BUDDY_LIST);
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL)
		return;

	if (new_val != imwindow->blist_visible)
		blist_toggle(priv->blist, screen.cur_window->data);
}

static void wopt_changed_timestamp(struct imwindow *imwindow) {
	swindow_set_timestamp(&imwindow->swindow,
		wopt_get_bool(imwindow->opts, WOPT_TIMESTAMP));

	imwindow_refresh(imwindow);
}

void wopt_init(struct window_opts *wopt, const char *target) {
	char nnick[NUSER_LEN];
	char buf[NUSER_LEN + 256];
	char *pork_dir;

	memset(wopt, 0, sizeof(struct window_opts) * WOPT_NUM_OPTS);

	wopt[WOPT_BEEP_ON_OUTPUT].val.b = opt_get_bool(OPT_BEEP_ON_OUTPUT);
	wopt[WOPT_HISTORY_LEN].val.i = opt_get_int(OPT_HISTORY_LEN);
	wopt[WOPT_LOG].val.b = opt_get_bool(OPT_LOG);
	wopt[WOPT_PRIVATE_INPUT].val.b = opt_get_bool(OPT_PRIVATE_INPUT);
	wopt[WOPT_SCROLLBUF_LEN].val.i = opt_get_int(OPT_SCROLLBUF_LEN);
	wopt[WOPT_SCROLL_ON_INPUT].val.b = opt_get_bool(OPT_SCROLL_ON_INPUT);
	wopt[WOPT_SCROLL_ON_OUTPUT].val.b = opt_get_bool(OPT_SCROLL_ON_OUTPUT);
	wopt[WOPT_SHOW_BUDDY_LIST].val.b = opt_get_bool(OPT_SHOW_BUDDY_LIST);
	wopt[WOPT_TIMESTAMP].val.b = opt_get_bool(OPT_TIMESTAMP);

	normalize(target, nnick, sizeof(nnick));

	pork_dir = opt_get_str(OPT_PORK_DIR);
	snprintf(buf, sizeof(buf), "%s/logs/%s.log", pork_dir, nnick);

	wopt[WOPT_LOGFILE].val.s = xstrdup(buf);
}

static int opt_tristate(char *args) {
	if (args == NULL)
		return (-1);

	if (!strcasecmp(args, "ON") ||
		!strcasecmp(args, "TRUE") ||
		!strcasecmp(args, "1"))
	{
		return (1);
	}

	if (!strcasecmp(args, "OFF") ||
		!strcasecmp(args, "FALSE") ||
		!strcasecmp(args, "0"))
	{
		return (0);
	}

	if (!strcasecmp(args, "TOGGLE"))
		return (2);

	return (-1);
}

static int opt_set_bool(u_int32_t opt, char *args) {
	int val = opt_tristate(args);

	if (val == -1)
		return (-1);

	if (val != 2)
		global_var[opt].val.b = val;
	else
		global_var[opt].val.b = !global_var[opt].val.b;

	if (global_var[opt].updated != NULL)
		global_var[opt].updated();

	return (0);
}

static int opt_set_char(u_int32_t opt, char *args) {
	if (args == NULL || *args == '\0')
		return (-1);

	global_var[opt].val.c = *args;

	if (global_var[opt].updated != NULL)
		global_var[opt].updated();

	return (0);
}

static int opt_set_int(u_int32_t opt, char *args) {
	u_int32_t num;
	char *end;

	if (args == NULL)
		return (-1);

	num = strtoul(args, &end, 0);
	if (*end != '\0')
		return (-1);

	global_var[opt].val.i = num;

	if (global_var[opt].updated != NULL)
		global_var[opt].updated();

	return (0);
}

static int opt_set_str(u_int32_t opt, char *args) {
	if (global_var[opt].dynamic == 1)
		free(global_var[opt].val.s);

	if (args != NULL) {
		global_var[opt].val.s = xstrdup(args);
		global_var[opt].dynamic = 1;
	} else {
		global_var[opt].val.s = NULL;
		global_var[opt].dynamic = 0;
	}

	if (global_var[opt].updated != NULL)
		global_var[opt].updated();

	return (0);
}

int opt_set(u_int32_t opt, char *args) {
	struct global_var *var = &global_var[opt];

	return (var->set(opt, args));
}

static int wopt_set_bool(struct imwindow *imwindow, u_int32_t opt, char *args) {
	int val = opt_tristate(args);

	if (val == -1)
		return (-1);

	if (val != 2)
		imwindow->opts[opt].val.b = val;
	else
		imwindow->opts[opt].val.b = !imwindow->opts[opt].val.b;

	if (window_var[opt].updated != NULL)
		window_var[opt].updated(imwindow);

	return (0);
}

static int wopt_set_char(struct imwindow *imwindow, u_int32_t opt, char *args) {
	if (args == NULL || *args == '\0')
		return (-1);

	imwindow->opts[opt].val.c = *args;

	if (window_var[opt].updated != NULL)
		window_var[opt].updated(imwindow);

	return (0);
}

static int wopt_set_int(struct imwindow *imwindow, u_int32_t opt, char *args) {
	u_int32_t num;
	char *end;

	if (args == NULL)
		return (-1);

	num = strtoul(args, &end, 0);
	if (*end != '\0')
		return (-1);

	imwindow->opts[opt].val.i = num;

	if (window_var[opt].updated != NULL)
		window_var[opt].updated(imwindow);

	return (0);
}

static int wopt_set_str(struct imwindow *imwindow, u_int32_t opt, char *args) {
	free(imwindow->opts[opt].val.s);

	if (args != NULL)
		imwindow->opts[opt].val.s = xstrdup(args);
	else
		imwindow->opts[opt].val.s = NULL;

	if (window_var[opt].updated != NULL)
		window_var[opt].updated(imwindow);

	return (0);
}

inline int wopt_set(struct imwindow *imwindow, u_int32_t opt, char *args) {
	struct window_var *var = &window_var[opt];

	return (var->set(imwindow, opt, args));
}
